//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// Class          : DiceEnsemble
//
// Author         : Richard E. Pattis
//                  Computer Science Department
//                  Carnegie Mellon University
//                  5000 Forbes Avenue
//                  Pittsburgh, PA 15213-3891
//                  e-mail: pattis@cs.cmu.edu
//
// Maintainer     : Author
//
//
// Description:
//
//   Objects constructed from the DiceEnsemble class act as collections of
// dice. The number of dice in an ensemble (and the number of sides of
// each die) can be controlled by the programmer.   
//
// Future Plans   : implement min/max methods
//
// Program History:
//   5/18/01: R. Pattis - Operational for 15-100
//   8/20/01: R. Pattis - JavaDoc comments added
//
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


package edu.cmu.cs.pattis.cs151xx;





/** 
 * Objects constructed from the <code>DiceEnsemble</code> class act
 *   as collections of dice.
 * The number of dice in an ensemble (and the number of sides of each die)
 *   can be controlled by the programmer.
 * The class models the basic operations need to roll the dice and determine
 *   the number of pips showing -both for individual dice and for the entire
 *   ensemble.   
 * 
 * @author Richard E. Pattis (Computer Science Department, Carnegie Mellon)
*/
public class DiceEnsemble { 


	/** 
	 * Constructs a <code>DiceEnsemble</code> object, specifying the number
	 *   of dice and the number of sides per die.
	 *
	 * @param  numberOfDice specifies the number of dice in the ensemble
	 * @param  sidesPerDie specifies the number of sides on each and every
	 *           die in the ensemble (the pips showing are 1, 2, 3, ...
	 *           <code>sidesPerDie</code>)
	 *
	 * @throws IllegalArgumentException if either parameter is less than 1
	*/
  public DiceEnsemble (int numberOfDice, int sidesPerDie)
    throws IllegalArgumentException
  {
    if (numberOfDice < 1)
      throw new IllegalArgumentException("DiceEnsemble constructor: Number of dice ("+numberOfDice+") < 1"); 
    if (sidesPerDie < 1)
      throw new IllegalArgumentException("DiceEnsemble constructor: Sides per die ("+sidesPerDie+") < 1"); 

    this.sidesPerDie  = sidesPerDie;
    this.pips         = new int[numberOfDice];
    //rollCount: see declaration for implicit initializaton to 0
  }
  
  
  
	/** 
	 * Constructs a <code>DiceEnsemble</code> object, consisting of
	 *   two dice, each having six sides: a standard ensemble.
	*/
  public DiceEnsemble ()
  {
    this(2,6);
  }
  
  
  
	/** 
	 * Returns this <code>DiceEnsemble</code> after rolling every die in it.
	 * By using the return type <code>DiceEnsemble</code> instead of
	 *  <code>void</code>, we can "cascade" method calls, writing expressions
	 *  such as: <code>d.roll().getPipSum()</code> instead of writing roll as
	 *  its own statement.
	 *
	 * @return the rolled dice ensemble
	*/
  public DiceEnsemble roll ()
  {
    rollCount++;
    for (int i=0; i<pips.length; i++)
      pips[i]  = (int)Math.round(Math.random()*sidesPerDie + .5); 
    return this;
  }

  
  
	/** 
	 * Returns the number of dice in this ensemble (set at construction).
	 *
	 * @return the number of dice in this ensemble
	*/
  public int getNumberOfDice ()
  {return pips.length;}
  
  
  
	/** 
	 * Returns the number of sides on each die in this ensemble (set at
	 *   construction).
	 *
	 * @return the number of sides on each die in this ensemble
	*/
  public int getSidesPerDie ()
  {return sidesPerDie;}
  
  

	/** 
	 * Returns the the number of pips showing on die <code>dieIndex</code>
	 *   (changed each time the ensemble is rolled).
	 *
	 * @param  dieIndex specifies the index of the die whose pips are returned
	 *
	 * @throws IllegalStateException if this ensemble has not been rolled
	 * @throws IllegalArguementException if <code>dieIndex</code> is outside
	 *           the range [1,getNumberOfDice()].
	 *
	 * @return the the number of pips showing on die <code>dieIndex</code>.
	*/
  public int getPips (int dieIndex)
     throws IllegalStateException, IllegalArgumentException
  {
    if (rollCount == 0)
      throw new IllegalStateException("getPip called, but dice not rolled");
    if (dieIndex < 1 || dieIndex > pips.length)
      throw new IllegalStateException("getPip called, but dieIndex("+dieIndex+") is not in the range [1,"+pips.length+"]");

    return pips[dieIndex-1];
  }
  
  

	/** 
	 * Returns the sum of all the pips showing in this ensemble (changed
	 *   each time the ensemble is rolled).
	 *
	 * @throws IllegalStateException if this ensemble has not been rolled
	 *
	 * @return the sum of all the pips showing in this ensemble
	*/
  public int getPipSum ()
    throws IllegalStateException
  {
    if (rollCount == 0)
      throw new IllegalStateException("getPipSum called, but dice not rolled");

    int pipSum = 0;
    for (int i=0; i<pips.length; i++)
      pipSum += pips[i];
    return pipSum;
  }
  
  

	/** 
	 * Returns the number of times the dice in this ensemble have been
	 *   rolled (incremented by 1 each time the ensemble is rolled).
	 *
	 * @return the number of times the dice in this ensemble have been
	 *         rolled
	*/
  public int getRollCount ()
  {return rollCount;}
  
  

	/** 
	 * Returns whether the pips showing for each die are all the same.
	 *
	 * @throws IllegalStateException if this ensemble has not been rolled
	 *
	 * @return whether the pips showing for each die are all the same;
	 *         for games like Monopoly, which use two dice, it is sometimes
	 *         important to know whether the user rolled doubles.
	*/
  public boolean allSame ()
    throws IllegalStateException
  {
    if (rollCount == 0)
      throw new IllegalStateException("allSame called, but dice not rolled");
    
    int unique = pips[0];
    for (int i=1; i<pips.length; i++)
      if (unique != pips[i])
        return false; 
    return true;
  }
  
  
  
	/** 
	 * Returns a <code>String<code> representation of the state of this ensemble.
	 *
	 * @return (1) the number of sides on each die, (2) the number of times the dice
	 *         ensemble has been rolled, and (3) the pips for each die (from which
	 *         the number of dice, their pip sum, and whether they all are the same
	 *         can be computed) 
	*/
  public String toString ()
  {
    String result =
       "DiceEnsemble[sidesPerDie="+getSidesPerDie()+","+
        "rollCount="+getRollCount()+","+
        "pips=[";
        
    for (int i=0; i<pips.length; i++)
      result += pips[i]+(i<pips.length-1?",":"]]");
      
    return result;
  }
  
  
  
  //No JavaDoc comments for private instance variables; they are well-named
  
  private int   sidesPerDie;
  private int   rollCount;
  private int[] pips;
  
  //Note: numberOfDice is not stored explicitly, but implicitly as pips.length
}




