//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// Class          : SimpleQueue
//
// Author         : Richard E. Pattis
//                  Computer Science Department
//                  Carnegie Mellon University
//                  5000 Forbes Avenue
//                  Pittsburgh, PA 15213-3891
//                  e-mail: pattis@cs.cmu.edu
//
// Maintainer     : Author
//
//
// Description:
//
//   Objects constructed from the SimpleQueue class act as collections of
// any objects. We can enqueue objects into this collection and dequeue
// objects from the collection (according to the First-In/First-Out
// ordering). Note that NO METHODS are called on these objects inside
// this class: we just move references around. 
//
// Future Plans   : None
//
// Program History:
//   9/25/01: R. Pattis - Operational for 15-100
//   8/10/04: R. Pattis - Javadoc comments added (added to course library)
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


package edu.cmu.cs.pattis.cs151xx;



/**
 * Objects constructed from the SimpleQueue class act as collections of
 *   any references to objects.
 * We can enqueue reference to objects into this collection and dequeue
 *   references to objects from the collection (according to the
 *   Firt-In/First-Out ordering).
 * Note that NO METHODS are called on these objects inside this class:
 *   we just move references around.
*/ 
public class SimpleQueue  { 


	/** 
	 * Constructs a <code>SimpleQueue</code> object, specifying the
	 *   initial size for it (number of references it can store).
	 * Initially, the <code>SimpleQueue</code> is empty.
	 *
	 * @param  initialSize specifies the number of references that the
	 *           <code>SimpleQueue</code> can initially store.
	 *"
	 * @throws IllegalArgumentException if <code>initialSize</b> is &lt 1
	*/
  public SimpleQueue (int initialSize)
    throws IllegalArgumentException
  {
    if (initialSize < 1)
      throw new IllegalArgumentException("SimpleQueue Constructor: initialSize("+initialSize+") < 1");

    q = new Object[initialSize];
  }
  
  
  
	/** 
	 * Constructs a <code>SimpleQueue</code> object, specifying the
	 *   initial size for it as 1.
	*/
  public SimpleQueue ()
  {this(1);}
  
  
  
  
  
	/** 
	 * Make this <code>SimpleQueue</code> empty.
	*/
  public void makeEmpty ()
  {
    for (int i=0; i<=rear; i++)
      q[i] = null;
    rear  = -1;
  }
  
  
  
	/** 
	 * Enqueue a reference onto the rear of this <code>SimpleQueue</code>.
	 *
	 * @param o specifies the reference to enqueue
	*/
  public void enqueue (Object o)
  {
    if ( getSize() == q.length)
        doubleLength();
           
    rear++;
    q[rear] = o;        //or just q[++rear] = o;
  }
  
  
  
	/** 
	 * Returns the front reference in this <code>SimpleQueue</code>, 
	 *   removing it in the process.
	 *
	 * @return the front reference on this <code>SimpleQueue</code>,
	 *   removing it in the process
	 *
	 * @throws IllegalStateException if this <code>SimpleQueue</code> is
	 *          empty
	*/
  public Object dequeue () throws IllegalStateException
  {
    if ( isEmpty() )
      throw new IllegalStateException("SimpleQueue dequeue: queue is empty");
    
    Object answer = q[0];
    
    //Shift all remaining values (q[1..rear]) left by 1 index, into
    //  positions  q[0..rear-1]
    for (int i=0; i<rear; i++)
      q[i] = q[i+1];
   
    //Remove duplicate in q[rear]; it has been copied into
    //  q[rear-1] too 
    q[rear] = null;
    rear--;                 //or just q[rear--] = null;
    return answer;
  }
    
    
  
  
	/** 
	 * Returns the front reference on this <code>SimpleQueue</code>, 
	 *   without removing it in the process.
	 *
	 * @return the front reference on this <code>SimpleQueue</code>,
	 *   without removing it in the process
	 *
	 * @throws IllegalStateException if this <code>SimpleQueue</code> is
	 *          empty
	*/
  public Object peek ()
    throws IllegalStateException
  {
    if ( isEmpty() )
      throw new IllegalStateException("SimpleQueue peek: queue is empty");
    
    return q[0];  
  }
  
  
  
	/** 
	 * Returns whether or not this <code>SimpleQueue</code> is empty
	 *   (stores no references).
	 *
	 * @return whether or not this <code>SimpleQueue</code> is empty
	 *   (stores no references)
	*/
  public boolean isEmpty ()
  {return rear == -1;}

  
  
	/** 
	 * Returns the number of reference this <code>SimpleQueue</code>
	 *   stores (0 if empty).
	 *
	 * @return the number of reference this <code>SimpleQueue</code>
	 *   stores (0 if empty)
	*/
  public int getSize ()
  {return rear+1;}

  
  
	/** 
	 * Returns a <code>String<code> representation of the state of this 
	 *   <code>SimpleQueue</code>.
	 * The <code>String</code> contains the value of <code>rear</code>,
	 *   the length of the <code>SimpleQueue</code> array, and all the references
	 *   stored in the <code>q</code> array. 
	 *
	 * @return  a <code>String<code> representation of the state of this 
	 *  <code>SimpleQueue</code>
	*/
  public String toString ()
  {
    String answer = "SimpleQueue[rear="+rear+"/length="+q.length;
    for (int i=0; i<=rear; i++)
      answer += ";q["+i+"]="+q[i];
    answer +=  "]";
    
    return answer;
  }
  
  


  
	/** 
	 * Doubles the length of the <code>q</code> array, while keeping
	 *   the same references in the same positions in it.
	 * This method is called only in <code>enqueue</code> when the array is
	 *   filled.
	*/
  private void doubleLength ()
  {
    Object[] temp = new Object[q.length*2];
    for (int i=0; i<q.length; i++)
      temp[i] = q[i];
    q = temp;
  }




  
	/** 
	 * Stores the array of references between indexes <code>0</code> and
	 *   <code>rear</code>.
	 * Its length is doubled when necessary, is always &gt <code>rear</code>,
	 *   and never shrinks.
	*/
  private Object[] q;
  
  
	/** 
	 * Stores the index of the rear of the queue.
	 * In an empty queue it stores -1.
	 * There are always <code>rear+1</code> references in the queue.
	*/
  private int      rear = -1;
}
