//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
//Program        : Closest-Points Tester
//
//Author         : Richard E. Pattis
//								 Computer Science Department
//								 Carnegie Mellon University
//								 5000 Forbes Avenue
//								 Pittsburgh, PA 15213-3891
//								 e-mail: pattis@cs.cmu.edu
//
//Maintainer     : Author
//
//
//Description:
//
//Test three closest-point algorithms for correctness 
//(well, at least agreement with each other's results) and speed.
//
//a) Brute Force computes the distance from every point to every other point,
//   keeping track of the smallest distance (and the two points generating it).
//   It is O(N^2)
//
//b) Sorting sorts an array of points first by their X and then by their Y 
//   coordinates (two O(NlogN) operations) and then recursively computes the 
//   closest points in two equal-sized half planes, and then combines these 
//   solutions by computing "cross-plane" distances (distances between points
//   in different half planes), subject to an important theorem limiting which 
//   points mut be examined.
//   It is O(NlogN)
//
//c) Hashing divides the unit square (normalize the data first) into subsquares
//   whose side lengths are based on the minimum computed distance so far (so
//   everything is rehashed when a new closest pair is found).  When a new point
//   is introduced, it is checked against at most 25 points near it: only those
//   that could possibly decrease the minimum distance.
//   It is O(N), although emperically the Sorting method always runs faster.
//
//For discussions of algorithms b and c, see Kleinberg and Tardos, "Algorithm
//Design", Prentice Hall 2005, Sections 5.4 (pg. 225-231); 13.7 (pg. 741-748)
// 
//
//Known Bugs     : None
//
//Future Plans   : None
//
//Program History:
//1/08/05: R. Pattis - Operational in Java for 15-200
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


import edu.cmu.cs.pattis.cs151xx.Prompt;
import edu.cmu.cs.pattis.cs151xx.Timer;

import java.awt.Point;
import java.awt.geom.Point2D;

import java.util.Arrays;
import java.util.Comparator;
import java.util.Map;
import java.util.HashMap;



public class Application {
	
	//For sorting Point2d.Double[] by their X and Y coordinates
	
  private static Comparator byX =
		new Comparator() {
		public int compare (Object o1, Object o2) {
			Point2D.Double p1 = (Point2D.Double)o1;
			Point2D.Double p2 = (Point2D.Double)o2;
			double diff = p1.getX()-p2.getX();
			if (diff < 0) return -1;
			if (diff ==0) return  0;
			return +1;
		}};
		
		private static Comparator byY =
			new Comparator() {
			public int compare (Object o1, Object o2) {
				Point2D.Double p1 = (Point2D.Double)o1;
				Point2D.Double p2 = (Point2D.Double)o2;
				double diff = p1.getY()-p2.getY();
				if (diff < 0) return -1;
				if (diff ==0) return  0;
				return +1;
			}};
			
			
			
  //Brute Force Algorithm
	//Note by starting j=i+1, we will compare each pair of points only once,
	//  and never compare a point against itself (that is, not compare point
	//  1 against point 2 and then later point 2 agains point 1)
	//The returned result is always an array of two points (the two closest).
  static Point2D.Double[] bruteFindClosest(Point2D.Double[] a) {
  	int i1 = -1, i2 = -1;
  	double minDistance = Double.MAX_VALUE, dist;
  	for (int i=0; i<a.length; i++)
  		for (int j=i+1; j<a.length; j++)
  			if ( (dist = a[i].distance(a[j])) < minDistance ) {
  				i1 = i;
  				i2 = j;
  				minDistance = dist;
  			}
  	
  	return new Point2D.Double[]{a[i1],a[i2]};
  }
  

  
  //Sorting Algorithm: a method and its recursive helper
  //Recursive helper
  //Optimize the value n for the "base case": |P| <= n
  static Point2D.Double[] closestPair(Point2D.Double[] pX,
  																		Point2D.Double[] pY) {
    //|P| <= 5, sort by brute force
  	if (pX.length <= 5)
      return bruteFindClosest(pX);
   	
  	//Compute Qx, Qy, Rx, Ry  in O(N) since parameters are sorted
  	//Q is left plane, R is right plane (following book: why not L, R?)
    int cutPoint = (pX.length+1)/2;  //If asymmetric, Q includes extra point
    Point2D.Double[] qX,qY,rX,rY;
    qX = new Point2D.Double[cutPoint];
    qY = new Point2D.Double[cutPoint];
    rX = new Point2D.Double[pX.length-cutPoint];
    rY = new Point2D.Double[pX.length-cutPoint];
    
    //Get the sorted points for this these planes
  	for (int i=0; i<cutPoint; i++) {
  		qX[i] = pX[i];
  		qY[i] = pY[i];
  	}
  	for (int i=cutPoint; i<pX.length; i++) {
  		rX[i-cutPoint] = pX[i];
  		rY[i-cutPoint] = pY[i];
  	}
  	
  	//Compute recursive solutions for Q and R
  	Point2D.Double[] qClosest = closestPair(qX,qY);	
  	Point2D.Double[] rClosest = closestPair(rX,rY);	
  	
  	//Compute minimum distance for Q/R crossovers
  	//Note: computes more than cross-overs, since some may be in same quadrant,
  	//      but still linearly bounded
  	double delta  = Math.min(qClosest[0].distance(qClosest[1]), rClosest[0].distance(rClosest[1]));
  	double xLine = qX[cutPoint-1].getX();
  	Point2D.Double[] sY = new Point2D.Double[pY.length];
    int sYLength = 0;
   	for (int i=0; i<pY.length; i++)
   		if (Math.abs(pY[i].getX()-xLine) < delta)
   			sY[sYLength++] = pY[i];
   	
   	int i1 = -1, i2 = -1;
   	double minDist = Double.MAX_VALUE, dist;
  	for (int i=0; i<sYLength; i++) 
  		for (int j=i+1; j<i+16 && j<sYLength; j++) 
 			  if ( (dist = sY[i].distance(sY[j])) < minDist ) {
  				i1 = i;
  				i2 = j;
  				minDist = dist;
  			}
  	
  	//Return result based on Q, R, and minimum crossover
    if (minDist < delta)
    	return new Point2D.Double[]{sY[i1],sY[i2]};
    else if (qClosest[0].distance(qClosest[1]) < rClosest[0].distance(rClosest[1]) )
    	return qClosest;
    else
    	return rClosest;
    	
  }
  
  
  //Sorting Algorithm
	//The returned result is always an array of two points (the two closest).
  static Point2D.Double[] sortedFindClosest(Point2D.Double[] p) {
  	//Sort these arrays once, here, and use them many times in closestPair
    Point2D.Double[] pX,pY;
    pX = new Point2D.Double[p.length];
    pY = new Point2D.Double[p.length];
    for (int i=0; i<p.length; i++)
  		pX[i] = pY[i] = p[i];
  	Arrays.sort(pX, byX);
 	  Arrays.sort(pY, byY);
	
 	  //Call helper with sorted arrays.
 	  return closestPair(pX,pY);
 }

	
  
  //Hash Algorithm: a method and its helper
  //Rehash Map m with square size delta/2 by adding length points from p
   static void rehash(Map m, double delta, Point2D.Double[] p, int length) {
   	 m.clear();
  	 for (int i=0; i<length; i++) {
  		 int s = (int)(p[i].getX()/(2*delta));
  		 int t = (int)(p[i].getY()/(2*delta));
  		 m.put(new Point(s,t), p[i]);
  	 }
   }
   
   
   
 	//The returned result is always an array of two points (the two closest).
   static Point2D.Double[] hashFindClosest(Point2D.Double[] p) {
     //Normalize points to be in [0,1)
     HashMap m = new HashMap();
     
     //Setup with first two points
     Point2D.Double p1 = p[0], p2 = p[1], pStar;
     double delta = p1.distance(p2), dist;
     int s,t;
     boolean changed;
     rehash(m,delta,p,2);
 
     //Add the rest of the points, searching the specified 5x5 neighborhood
     //  for closer points, either rehashing for a new delta (if found) or
     //  just adding the point to the current hash
     for (int i=2; i<p.length; i++) {
 		   s = (int)(p[i].getX()/(2*delta));
  		 t = (int)(p[i].getY()/(2*delta));
  		 changed = false;
  		 for (int j=s-2; j<=s+2; j++)
  	 		 for (int k=t-2; k<=t+2; k++) {
  	 			 pStar = (Point2D.Double)m.get(new Point(j,k));
           if (pStar != null && (dist = pStar.distance(p[i])) < delta) {
          	 changed = true;
          	 p1 = pStar;
          	 p2 = p[i];
          	 delta = dist;
           }
  	 		 }
  		   if (changed)
  		  	 rehash(m,delta,p,i+1);
  		   else
  		  	 m.put(new Point(s,t), p[i]);
  		}
     
      //Return last points used to calculate delta
     	return new Point2D.Double[]{p1,p2};	
   }
    
    
   
    //////////////////////////////
    //
    // main Method: driver
    //
    //////////////////////////////
    public static void main(String[] args) {
		
		for(;;) {
			char toTest = Prompt.forChar("Enter b)rute, h)ash, or s)ort","bhs",'h');
			int  n      = Prompt.forInt("Enter # of Points");
			int  diffs  = Prompt.forInt("Enter # of test arrays",5);
			int  times  = Prompt.forInt("Enter # of times to test each array",5);


			Point2D.Double[] a = new Point2D.Double[n];
			
			Timer t = new Timer();
			Point2D.Double[] answer = null;
			for (int diff=1; diff<=diffs; diff++) {
		  	for (int i=0; i<n; i++)
          a[i] = new Point2D.Double(Math.random(),Math.random());
		  	
				t.start();
		  	for (int time=1; time<=times; time++) {
	 				if (toTest == 'h') answer = hashFindClosest(a);
	 				if (toTest == 'b') answer = bruteFindClosest(a);
  				if (toTest == 's') answer = sortedFindClosest(a);
  			}
				t.stop();
		  }

			if (Prompt.forBoolean("\nDisplay Points",n<=50)) {
				for (int i=0; i<n; i++)
	        System.out.println(a[i]);
			}
	
			if ( Prompt.forBoolean("\nCalculate via brute (for checking)", toTest!='b' && n>=50) ) {
			  Point2D.Double[] ref = bruteFindClosest(a);
			  System.out.println("\nReference Points are\n" + ref[0] + "\n and " + ref[1]);
			}
			System.out.println("\nClosest Points are\n" + answer[0] + "\n and " + answer[1]);
			System.out.println("\nAverage Time = " + t.getElapsed()/(diffs*times) + " seconds");
	
			if (!Prompt.forBoolean("\nTest again",true))
				 break;
		}
	}

}
